# -*- coding: utf-8 -*-
"""
Functions related to modelling of energy assets.

This file can also be imported as a module and contains the following
functions:

    * wind_turbine_generator - converts wind speed to power output of a wind turbine
    * solar_pv_generator - convert solar irradiance to the power output of a PV installation


References:
    J. Twidell & T. Weir, Renewable Energy Resources, 2nd edition, 
    Taylor and Francis, 2006

"""

# Import all useful libraries
import casadi as ca
import casadi.tools
import numpy as np


def wind_turbine_generator(
    wind_speed,
    rated_power=1000,
    cut_in_wind_speed=5,
    rated_wind_speed=12,
    cut_out_wind_speed=30,
):
    """Models the power production curve of a standard wind turbine.
    The model is based upon (Twidell, 2006), pp. 306-307.

    Parameters
    ----------
    wind_speed : np.array
        Wind speed [m/s]
    rated_power : float
        Rated power of the turbine [W]
        (default: 1000)
    cut_in_wind_speed : float
        Cut-in wind speed of the turbine [m/s]
        (default: 5)
    rated_wind_speed : float
        Rated wind speed of the turbine [m/s]
        (default: 12)
    cut_out_wind_speed : float
        Cut-out wind speed of the turbine [m/s]
        (default: 30)

    Raises
    ------
    None

    Returns
    -------
    power_output : np.array
        Power production corresponding to the values of wind_speed [W]
    """

    k_cubic = np.where(
        (wind_speed >= cut_in_wind_speed) & (wind_speed < rated_wind_speed)
    )
    k_rated = np.where(
        (wind_speed >= rated_wind_speed) & (wind_speed <= cut_out_wind_speed)
    )

    # k_zero = np.where((wind_speed<=cut_in_wind_speed) | (wind_speed>cut_out_wind_speed))
    power_output = np.zeros(np.size(wind_speed))

    u_ci_3 = cut_in_wind_speed ** 3
    u_R_3 = rated_wind_speed ** 3
    a = rated_power / (u_R_3 - u_ci_3)
    b = u_ci_3 / (u_R_3 - u_ci_3)
    power_output[k_cubic,] = (
        a
        * (
            wind_speed[
                k_cubic,
            ]
            ** 3
        )
        - b * rated_power
    )

    power_output[
        k_rated,
    ] = rated_power

    return power_output


def solar_pv_generator(solar_radiation, rated_power=1000, rated_radiation=1000):
    """Models the power production curve of a standard solar PV panel.
    The model is based upon a simplified linear model.

    Parameters
    ----------
    wind_speed : np.array
        Values of the incident radiation on the panel [W/m^2]
    rated_power : float
        Rated power of the inverter of the PV panel [W]
        (default: 1000)
    rated_radiation : float
        Values of the design radiation on the panel [W/m^2]
        (default: 1000)

    Raises
    ------
    None

    Returns
    -------
    power_output : np.array
        Power production corresponding to the values of solar_radiation [W]
    """
    power_output = rated_power * (solar_radiation / rated_radiation)

    # Forcing the output to not exceed the rated power
    k_excess = np.where(power_output > rated_power)
    power_output[
        k_excess,
    ] = rated_power

    return power_output


def battery_optimal_controller(
    price,
    local_demand=None,
    local_production=None,
    initial_soc=0,
    capacity=1e4,
    rated_power=1e3,
    charge_efficiency=0.98,
    discharge_efficiency=0.98,
    transformer_sizing=np.inf,
):

    ## System description
    # State vector
    x = ca.tools.struct_symMX(["SOC"])

    # Input vector
    u = ca.tools.struct_symMX(["Pin", "Pout"])

    # State equations
    dxdt = ca.tools.struct_MX(x)

    dxdt["SOC"] = charge_efficiency * u["Pin"] - u["Pout"] / discharge_efficiency

    # ODE Right-hand side
    rhs = dxdt
    f = ca.Function("f", [x, u], [dxdt], ["x", "u"], ["dx/dt"])

    dt = ca.MX.sym("dt")

    # RK4
    k1 = f(x, u)
    k2 = f(x + dt / 2.0 * k1, u)
    k3 = f(x + dt / 2.0 * k2, u)
    k4 = f(x + dt * k3, u)
    xf = x + dt / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4)

    # Single step time propagation
    F_RK4 = ca.Function("F_RK4", [x, u, dt], [xf], ["x[k]", "u[k]", "dt"], ["x[k+1]"])

    ## ----------- Optimal control -------------------
    opti = ca.Opti()

    # Optimization horizon
    N = len(price)
    price = price.reshape(1, N)

    # Decision variables for states and inputs
    X = opti.variable(x.size, N + 1)

    SOC = X[0, :]
    U = opti.variable(u.size, N)

    if local_demand is None and local_production is None:
        y_Pimport = U[0, :]
        y_Pexport = U[1, :]
    else:
        if local_demand is None:
            local_demand4opt = np.zeros(N)
        else:
            local_demand4opt = local_demand

        if local_production is None:
            local_production4opt = np.zeros(N)
        else:
            local_production4opt = local_production

        local_demand4opt = local_demand4opt.reshape(1, N)
        local_production4opt = local_production4opt.reshape(1, N)

        y_Pimport = ca.fmax(
            0, U[0, :] + local_demand4opt - U[1, :] - local_production4opt
        )
        y_Pexport = ca.fmax(
            0, U[1, :] + local_production4opt - U[0, :] - local_demand4opt
        )

    Pin_battery = U[0, :]
    Pout_battery = U[1, :]

    # Initial state is a parameter
    x0 = initial_soc
    opti.subject_to(X[:, 0] == x0)
    opti.subject_to(X[0, :] <= capacity)
    opti.subject_to(X[0, :] >= 0)

    opti.subject_to(U[0, :] <= rated_power)
    opti.subject_to(U[0, :] >= 0)
    opti.subject_to(U[1, :] <= rated_power)
    opti.subject_to(U[1, :] >= 0)

    # State constraints
    for k in range(N):
        opti.subject_to(X[:, k + 1] == F_RK4(X[:, k], U[:, k], 1))

    if np.isinf(transformer_sizing) is False:
        opti.subject_to(y_Pimport <= transformer_sizing)
        opti.subject_to(y_Pexport <= transformer_sizing)

    revenue = -ca.sum2(ca.times(y_Pimport, price)) + ca.sum2(ca.times(y_Pexport, price))
    opti.minimize(-revenue)

    opti.solver("ipopt")
    sol = opti.solve()

    t = np.linspace(1, N, num=N, axis=0)
    val_Pout = sol.value(Pout_battery)
    val_Pin = sol.value(Pin_battery)
    val_SOC = sol.value(SOC)
    val_revenue = sol.value(revenue)
    val_export = sol.value(y_Pexport)
    val_import = sol.value(y_Pimport)

    performance = dict(
        revenue=np.sum(val_revenue),
        energy_output=np.sum(val_Pout),
        energy_intake=np.sum(val_Pin),
        energy_losses=np.sum(val_Pin) - np.sum(val_Pout),
        energy_export=np.sum(val_export),
        energy_import=np.sum(val_import),
    )

    timeseries = dict(
        t=t,
        soc=val_SOC,
        input_power=val_Pin,
        output_power=val_Pout,
        revenue=val_revenue,
        price=price.reshape(N, 1),
        energy_export=val_export,
        energy_import=val_import,
    )

    timeseries["accumulated_revenue"] = np.cumsum(
        np.multiply(
            (timeseries["energy_export"] - timeseries["energy_import"]).reshape(1, N),
            timeseries["price"].reshape(1, N),
        )
    )

    if local_demand is None:
        performance["local_demand"] = np.sum(local_demand)
        timeseries["local_demand"] = local_demand
    if local_production is None:
        performance["local_production"] = np.sum(local_production)
        timeseries["local_production"] = local_production

    return performance, timeseries



class MissingApiKeyError(Exception):
    """Raised when an API key has not been loaded.

    Attributes:
        message -- explanation of why the specific transition is not allowed
        environment_variable -- enviroment variable that should contain the key
    """

    def __init__(self, message, environment_variable, create_at_url):
        print(
            "You are missing an API loaded in the environment variable "
            + environment_variable
            + ".\nIf you do not have one, create one on: "
            + create_at_url
            + "\n\n"
        )
        self.message = message
        self.environment_variable = environment_variable
        self.create_at_url = create_at_url



# -*- coding: utf-8 -*-
"""
Functions related to data loading in Denmark.

API specification is found on:
    https://confluence.govcloud.dk/pages/viewpage.action?pageId=15303111

functions:

    * get_weather_stations - loads the list of all Danish weather stations
    * get_weather_station_data - loads historical data for a given station
    * get_closest_weather_station - identifies the closest weather station
    * get_data_for_coordinates - loads historical data from the closest station
        to a set of coordinates

"""

# Import all useful libraries
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import os

from seedftw.exceptions import MissingApiKeyError
from seedftw.environment.geography import distance_between_coordinates

# For caching
from functools import lru_cache


def _get_api_key():
    api_key_field = "SEEDFTW:API_KEY:DK:DMI"
    dmi_api_key = os.getenv(api_key_field)
    if dmi_api_key is None:
        e = MissingApiKeyError(
            message="Missing API key for DMI service.",
            environment_variable=api_key_field,
            create_at_url="https://confluence.govcloud.dk/display/FDAPI",
        )
        raise e

    return dmi_api_key


# Functions dealing with microepoch data
from seedftw.base.timeseries import (
    format_to_timetable,
    datetime_to_microepoch,
    microepoch_to_datetime_index,
    microepoch_to_local_datetime,
    synchronise,
    split_data_loading_range,
)


def _datetime_to_microepoch(datetimes2use):
    return datetime_to_microepoch(datetimes2use)


def _microepoch_to_datetime_index(microepoch):
    return microepoch_to_datetime_index(microepoch)


def _microepoch_to_local_datetime(microepoch):
    return microepoch_to_local_datetime(microepoch)


def __format_data_table(data):
    return format_to_timetable(data, time_column="Minutes5UTC", column_dict=None)


# Loading data from DMI's API
@lru_cache(maxsize=10)
def __dmi_api_call(address):

    api_key_part = "api-key=" + _get_api_key()

    query_url = "https://dmigw.govcloud.dk/" + address
    if "?" in query_url:
        binding_sym = "&"
    else:
        binding_sym = "?"

    ur_2_call = query_url + binding_sym + api_key_part

    # Read data in Json format
    result = pd.read_json(ur_2_call)

    return result


# Loading data from the API for 1 parameter
def __load_parameter_timeseries(station, parameter, start_date, end_date, resolution):
    """
    Documentation of the details of the data is available here:
    >>  https://confluence.govcloud.dk/pages/viewpage.action?pageId=26476616
    """

    if isinstance(station, str):
        station_string = station
    else:
        station_string = str(station).zfill(5)

    parameter2id = {
        "ambient_temperature": "temp_dry",
        "wind_speed": "wind_speed",
        "wind_direction": "wind_dir",
        "relative_humidity": "humidity",
        "global_solar_radiation": "radia_glob",
    }

    # Adding a query size limiter to approx. 6 months
    limit2use = 30000
    max_data_length = timedelta(weeks=24)
    date_steps_range = split_data_loading_range(
        start_date, end_date, step=max_data_length
    )
    data = None
    for start_j, end_j in date_steps_range:

        start_microepoch = _datetime_to_microepoch(start_j)
        end_microepoch = _datetime_to_microepoch(end_j - timedelta(seconds=1))
        query_url = "metObs/v1/observation?from={}&to={}&stationId={}&parameterId={}&limit={}".format(
            str(start_microepoch),
            str(end_microepoch),
            station_string,
            parameter2id[parameter],
            str(limit2use),
        )

        dmi_raw_data = __dmi_api_call(query_url)
        dmi_raw_data["t"] = _microepoch_to_datetime_index(dmi_raw_data["timeObserved"])
        data_j = format_to_timetable(
            dmi_raw_data[["t", "value"]],
            time_column="t",
            column_dict={"value": parameter},
        )
        if data is None:
            data = data_j
        else:
            data = data.append(data_j)

    if resolution == "raw" or data is None:
        # Do nothing
        None
    elif resolution == "hour":
        data = data.resample("H").mean()
    elif resolution == "day":
        # Averaging per day is made in local timezone
        data.index = data.index.tz_convert("Europe/Copenhagen")
        data = data.resample("D").mean()
        data.index = data.index.tz_convert("UTC")
    else:
        print("Error: invalid resolution:" + resolution)

    return data


@lru_cache(maxsize=1)
def get_weather_stations():
    """Loads the list of the Danish weather stations

    Parameters
    ----------
    None

    Raises
    ------
    None

    Returns
    -------
    stations : pandas.DataFrame(height,latitude,longitude,name,type,stationId)
        Details of the weather stations
    """
    T = __dmi_api_call("metObs/v1/station?country=DNK")

    stations = T["location"].apply(pd.Series)
    stations["name"] = T["name"]
    stations["type"] = T["type"]
    stations["stationId"] = T["stationId"]

    return stations


def get_weather_station_data(
    station=6031,
    start_date=(datetime.now() - timedelta(days=3)),
    end_date=datetime.now(),
    resolution="hour",
    parameters=["ambient_temperature"],
):
    """Loads historical data for a given weather station

    Parameters
    ----------
    station : int or str
        stationId of the station for which data is to be loaded
    start_date : datetime
        Time for the start of the historical period
    end_date: datetime
        Time for the end of the historical period
    resolution: str
        Resolution of the data to load (raw,hour,day)
    parameters : list(str)
        List of parameters to be loaded
        (Options: "ambient_temperature","wind_speed","wind_direction",
         "relative_humidity","global_solar_radiation")

    Raises
    ------
    None

    Returns
    -------
    data : pandas.DataFrame(t,ambient_temperature)
        Timetable of the ambient temperature [degrees C]


    Documentation of the details of the data is available here:
    >>  https://confluence.govcloud.dk/pages/viewpage.action?pageId=26476616

    """

    if isinstance(station, str):
        station_string = station
    else:
        station_string = str(station).zfill(5)

    data = None

    for par_i in parameters:
        data_i = __load_parameter_timeseries(
            station, par_i, start_date, end_date, resolution
        )
        if data is None:
            data = data_i
        else:
            data = synchronise(data, data_i, "union")

    return data


def get_closest_weather_station(latitude, longitude):
    """Identifies the closest weather station for a set geographical coordinates

    Parameters
    ----------
    latitude : float
        Latitude [degrees]
    longitude : float
        Longitude [degrees]

    Raises
    ------
    None

    Returns
    -------
    closest_station : pandas.Series(height,latitude,longitude,name,type,stationId)
        Details of the closest location
    """

    # Get all DMI locations
    dmi_stations = get_weather_stations()

    # Selecting the ones with all weather data available
    selected_stations = dmi_stations.loc[dmi_stations["type"] == "Synop", :]

    distance_to_stations = selected_stations.apply(
        lambda row: distance_between_coordinates(
            [latitude, longitude], [row["latitude"], row["longitude"]]
        ),
        axis=1,
    )

    closest_index = selected_stations.index[np.argmin(distance_to_stations)]
    closest_station = selected_stations.loc[closest_index, :]

    return closest_station


def get_data_for_coordinates(
    latitude=57.046707,
    longitude=9.935932,
    start_date=(datetime.now() - timedelta(days=3)),
    end_date=datetime.now(),
    resolution="hour",
    parameters=["ambient_temperature"],
):
    """Loads historical data from the closest weather station to a set of coordinates

    Parameters
    ----------
    latitude : float
        Latitude [degrees]
    longitude : float
        Longitude [degrees]
    start_date : datetime
        Time for the start of the historical period
    end_date: datetime
        Time for the end of the historical period
    resolution: str
        Resolution of the data to load (raw,hour,day)
    parameters : list(str)
        List of parameters to be loaded
        (Options: "ambient_temperature","wind_speed","wind_direction",
         "relative_humidity","global_solar_radiation")

    Raises
    ------
    None

    Returns
    -------
    data : pandas.DataFrame(t,ambient_temperature)
        Timetable of the ambient temperature [degrees C]
    """

    closest_station = get_closest_weather_station(latitude, longitude)
    data = get_weather_station_data(
        station=closest_station["stationId"],
        start_date=start_date,
        end_date=end_date,
        resolution=resolution,
        parameters=parameters,
    )
    return data
